from __future__ import annotations
from math import radians, degrees, sin, cos, sqrt, atan2


class Coordinates:
    earth_radius = 6_364_814  # in meters at Warsaw latitude

    def __init__(self, longtitude: float, latitude: float) -> Coordinates:
        self.__theta = radians(longtitude)
        self.__phi = radians(latitude)

    # The errors are too big
    # def distance(self, other : Coordinates) -> float:
    #   sqr = 2 * (1 -
    #               \ sin(self.__theta)* sin(other.__theta) * cos(self.__phi - other.__phi)
    #                \ - cos(self.__theta)*cos(other.__theta) )
    #   if (sqr < 0):
    #     sqr = 0

    #   return Coordinates.earth_radius * sqrt(sqr)

    def distance(self, other: Coordinates):
        dLat = self.__phi - other.__phi
        dLon = self.__theta - other.__theta
        a = sin(dLat/2)**2 + cos(self.__phi)*cos(other.__phi)*sin(dLon/2)**2
        c = 2 * atan2(sqrt(a), sqrt(1-a))
        return Coordinates.earth_radius * c

    def __sub__(self, other: Coordinates) -> Coordinates:
        return self.distance(other)

    @property
    def lon(self) -> float:
        return degrees(self.__theta)

    @property
    def lat(self) -> float:
        return degrees(self.__phi)

    def __str__(self) -> str:
        return f"Lon: {degrees(self.__theta)},  Lat: {degrees(self.__phi)}"

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: Coordinates) -> bool:
        return self.__theta == other.__theta and self.__phi == other.__phi



from __future__ import annotations
import requests
import json
import time
from timeit import default_timer as timer
import concurrent.futures
from typing import Dict, List, Tuple, Optional, DefaultDict
from concurrent.futures import as_completed
from datetime import datetime
from collections import defaultdict
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


class ZTMFetcher:
    def __init__(self, apikeys: List[str]) -> ZTMFetcher:
        self.__apikeys = apikeys
        self.__session = requests.Session()
        self.__session.mount(
            "https://", HTTPAdapter(max_retries=Retry(total=10, backoff_factor=0.5)))
        self.__number_of_requests: int = 0

        self.__bustram_endpoint = "https://api.um.warszawa.pl/api/action/busestrams_get/"
        self.__schedule_endpoint = "https://api.um.warszawa.pl/api/action/dbtimetable_get"
        self.__db_endpoint = "https://api.um.warszawa.pl/api/action/dbstore_get/"
        self.__routes_end_point = "https://api.um.warszawa.pl/api/action/public_transport_routes/?"

    def __send_request(self, endpoint: str, params_without_apikey: Dict[str, str]):
        time.sleep(0.01)
        self.__number_of_requests += 1
        apikey = self.__apikeys[(self.__number_of_requests//100) % len(self.__apikeys)]
        params_without_apikey['apikey'] = apikey
        response = self.__session.get(endpoint, params=params_without_apikey)
        return response

    def __log(self, error_msg: str):
        print(f"Wrong response: {error_msg}")

    def fetch_bus_location(self) -> Dict[str, Dict[str, str]]:
        params = {'resource_id': '%20f2e5503e%02927d-4ad3-9500-4ab9e55deb59', 'type': '1'}
        bus_data: Dict[str, Dict[str, str]] = {}
        while True:
            """ Sample respone
            {
              "result": [
                  {"Lines": "219", "Lon": 21.1128748, "VehicleNumber": "1000", 
                   "Time": "2024-02-04 21:19:45", "Lat": 52.213404,"Brigade": "3"
                  },
                  { ... },
                        ],
              },
            """
            response = self.__send_request(self.__bustram_endpoint, params)
            if not response:
                continue
            response_json = response.json()
            if not response_json['result']:
                raise RuntimeError('No Buses')
            try:
                for bus in response_json['result']:
                    date = bus['Time'].split(' ')[0]
                    if date == datetime.now().strftime('%Y-%m-%d'):
                        bus_data[bus['VehicleNumber']] = {
                            "Lines": bus["Lines"], "Lon": bus["Lon"], "Lat": bus["Lat"], 
                            "Time": bus["Time"], "Brigade": bus["Brigade"]}
                return bus_data
            except TypeError:
                self.__log(response_json)

    def fetch_bus_location_period(self, period_second: int, 
                                  output_filename: Optional[str] = None) -> None:
        bus_data: Dict[int, Dict[str, Dict[str, str]]] = {}
        i = 0
        times = period_second // 10
        start_time = datetime.now().strftime('%Y-%m-%d %H-%M-%S')
        for _ in range(times):
            bus_data[i] = self.fetch_bus_location()
            print(f"Collection {i} done")
            i += 1
            time.sleep(10)

        output = output_filename if output_filename is not None else \
            f"bus_coordinates_{start_time}_duration_{int(period_second)}s.json"

        with open(output, "w", encoding='utf-8') as fp:
            json.dump(bus_data, fp, ensure_ascii=False)

    def __fetch_busstop_coordinates_helper(self) -> Dict[str, Tuple[float, float]]:
        params = {'id': '1c08a38c-ae09-46d2-8926-4f9d25cb0630'}
        busstop_coord: Dict[str, Tuple[float, float]] = {}
        while True:
            """ Sample response:
            {
              "result": [
                          {
                            "values": [
                                        {"value": "1001", "key": "zespol"},
                                        {"value": "01", "key": "slupek"},
                                        {"value": "Kijowska","key": "nazwa_zespolu"},
                                        {"value": "2201","key": "id_ulicy"},
                                        {"value": "52.248455","key": "szer_geo"},
                                        {"value": "21.044827","key": "dlug_geo"},
                                        {"value": "al.Zieleniecka","key": "kierunek"},
                                        {"value": "2023-09-23 00:00:00.0","key": "obowiazuje_od"}
                                      ]
                          },
                        ]
            }
            """
            response = self.__send_request(self.__db_endpoint, params)
            if not response:
                continue
            response_json = response.json()
            if not response_json['result']:
                raise ValueError(f"Empty result, fetch_busstop_coordinates")
            try:
                for bus_stop_dict in response_json['result']:
                    complex_id = bus_stop_dict['values'][0]['value']
                    bus_stop_id = bus_stop_dict['values'][1]['value']
                    bus_stop_name = bus_stop_dict['values'][2]['value']
                    latitude = bus_stop_dict['values'][4]['value']
                    longtitude = bus_stop_dict['values'][5]['value']
                    direction = bus_stop_dict['values'][6]['value']

                    key = complex_id + ',' + bus_stop_id
                    busstop_coord[key] = {
                        "Lon,Lat": (float(longtitude), float(latitude)),
                        "stop_name": bus_stop_name
                    }

                break
            except TypeError:
                self.__log(response_json)

        return busstop_coord

    def fetch_busstop_coordinates(self, output_filename: Optional[str] = None):
        busstop_data = self.__fetch_busstop_coordinates_helper()
        output = output_filename if output_filename is not None else \
            f"busstop_coordinates_{datetime.now().strftime('%Y-%m-%d')}.json"

        with open(output, "w", encoding='utf-8') as fp:
            json.dump(busstop_data, fp, ensure_ascii=False)

    def fetch_lines_for_busstop(self, bus_stop_complex_id: str, bus_stop_id: str):
        params = {'id': '88cd555f-6f31-43ca-9de4-66c479ad5942',
                  'busstopId': bus_stop_complex_id,
                  'busstopNr': bus_stop_id,
                 }
        lines_for_stop = []
        while True:
            """Sample response
              {'result': 
                      [
                        {'values': [{'value': '138', 'key': 'linia'}]}, 
                        {'values': [{'value': '143', 'key': 'linia'}]}, 
                        {'values': [{'value': '523', 'key': 'linia'}]}, 
                        {'values': [{'value': '525', 'key': 'linia'}]}, 
                        {'values': [{'value': 'N25', 'key': 'linia'}]}
                      ]
              }
            """
            response = self.__send_request(self.__schedule_endpoint, params)
            if not response:
                continue
            response_json = response.json()

            if not response_json['result']:
                return []
            try:
                for line_dict in response_json['result']:
                    lines_for_stop.append(line_dict['values'][0]['value'])
                return lines_for_stop
            except TypeError:
                self.__log(response_json)

    def fetch_routes(self, output_filename: Optional[str] = None):
        result_dict = defaultdict(lambda: defaultdict(dict))
        while True:
            """ Sample response:
                  {
                    "result": {
                      "1": {
                          "TD-1NAR": 
                            {
                              "1": {"odleglosc": 0, "ulica_id": "1501", "nr_zespolu": "R-01",
                                    "typ": "6", "nr_przystanku": "00"},
                              "2": {...}
                            },
                  },
            """
            response = self.__send_request(self.__routes_end_point, dict())
            if not response:
                continue
            response_json = response.json()
            if not response_json['result']:
                raise ValueError(f"Empty result")
            try:
                for line, routes in response_json['result'].items():
                    for route, info_dict in routes.items():
                        for nr_porzadkowy, info in info_dict.items():
                            result_dict[line][route][int(
                                nr_porzadkowy)] = info["nr_zespolu"]+','+info["nr_przystanku"]
                break
            except TypeError:
                self.__log(response_json)

        output = output_filename if output_filename is not None else \
            f"routes_{datetime.now().strftime('%Y-%m-%d')}.json"

        with open(output, "w", encoding='utf-8') as fp:
            json.dump(result_dict, fp, ensure_ascii=False)

    def __fetch_schedule_for_busstop_line(self, timetable: DefaultDict[str, DefaultDict[str, List]],
                                          bus_stop_complex_id: str, bus_stop_id: str,
                                          line_num: str) -> None:
        params = {'id': 'e923fa0e-d96c-43f9-ae6e-60518c9f3238',
                  'busstopId': bus_stop_complex_id,
                  'busstopNr': bus_stop_id,
                  'line': line_num,
                 }
        while True:
            """Sample response:
              {'result': 
                      [
                        { 'values': 
                                    [
                                      {'value': 'null', 'key': 'symbol_2'}, 
                                      {'value': 'null', 'key': 'symbol_1'},
                                      {'value': '1', 'key': 'brygada'},
                                      {'value': 'Utrata', 'key': 'kierunek'},
                                      {'value': 'TP-UTS', 'key': 'trasa'},
                                      {'value': '05:06:00', 'key': 'czas'}
                                    ]
                        }, 
                        { 'values': [...]},
                      ]
              }
              """
            response = self.__send_request(self.__schedule_endpoint, params)
            if not response:
                continue
            response_json = response.json()
            if not response_json['result']:
                return

            try:
                for bus_dict in response_json['result']:
                    brigade = bus_dict['values'][2]['value']
                    route = bus_dict['values'][4]['value']
                    time = bus_dict['values'][5]['value']
                    time = datetime.now().strftime('%Y-%m-%d') + " " + time
                    timetable[line_num][brigade][','.join((bus_stop_complex_id, bus_stop_id))].append({
                        "Time": time, "Route": route})
                return
            except TypeError:
                self.__log(response_json)

    def __fetch_schedule_for_busstop(self, timetable: DefaultDict[str, DefaultDict[str, List]],
                                     bus_stop_complex_id: str, bus_stop_id: str) -> None:
        all_lines = self.fetch_lines_for_busstop(
            bus_stop_complex_id, bus_stop_id)
        if len(all_lines) >= 10:
            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = [executor.submit(self.__fetch_schedule_for_busstop_line,
                                           timetable, bus_stop_complex_id, bus_stop_id, line)
                           for line in all_lines]
                for future in as_completed(futures): future.result()
        else:
            for line in all_lines:
                self.__fetch_schedule_for_busstop_line(
                    timetable, bus_stop_complex_id, bus_stop_id, line)

    def read_the_whole_schedule(self, output_filename: Optional[str] = None):
        start = timer()
        timetable = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        busstop_data = self.__fetch_busstop_coordinates_helper()

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [executor.submit(self.__fetch_schedule_for_busstop, timetable, *val.split(','))
                       for val in busstop_data.keys()]

            for future in as_completed(futures):
                future.result()

        output = output_filename if output_filename is not None else \
            f"timetable_{datetime.now().strftime('%Y-%m-%d')}.json"

        with open(output, "w", encoding='utf-8') as fp:
            # encode dict into JSON
            json.dump(timetable, fp, ensure_ascii=False)

        print("Time passed: ", timer() - start, " seconds")





from __future__ import annotations
from typing import Dict, List
import json

from .time_changer import time_changer


class TimeTableMangager:
    def __init__(self, timetable_file: str) -> TimeTableMangager:
        with open(timetable_file, 'r', encoding='utf8') as file:
            self.__timetable: Dict[str, Dict[str, Dict[str,
                                                       List[Dict[str, str]]]]] = json.load(file)
        for brigade_busstom_time in self.__timetable.values():
            for busstop_time in brigade_busstom_time.values():
                for list_of_timetable in busstop_time.values():
                    for time_dict in list_of_timetable:
                        time_dict["Time"] = time_changer(time_dict["Time"])

    def get_schedule(self, line: str, brigade: str) -> Dict[str, List[Dict[str, str]]]:
        return self.__timetable[line][brigade]



from __future__ import annotations
from typing import Optional
from collections import defaultdict
import pandas as pd
import folium
from folium.plugins import MarkerCluster
import plotly_express as px
import webbrowser

class Visualizer:
    def __init__(self) -> Visualizer:
        self.__warsaw_lon = 21.017532
        self.__warsaw_lat = 52.237049

    def heat_map_speeds(self, stats: pd.DataFrame, html_out: Optional[str] = None):
        fig = px.density_mapbox(stats, lat='lat', lon='lon', z='speed', radius=12,
                                center={'lat': self.__warsaw_lat, 'lon': self.__warsaw_lon}, 
                                zoom=12, mapbox_style='open-street-map', height=900)
        fig.update_layout(margin=dict(b=0, t=0, l=0, r=0))
        if html_out is not None:
            fig.write_html(html_out)
        fig.show()

    def scatter_map_speeds(self, stats: pd.DataFrame, html_out: Optional[str] = None):
        fig = px.scatter_mapbox(stats, lat='lat', lon='lon', color='speed',
                                color_continuous_scale=[
                                    [0.0, "green"],
                                    [0.3125, "green"],
                                    [0.3125111111, "yellow"],
                                    [0.53125, "yellow"],
                                    [0.5312511111, "orange"],
                                    [0.75, "orange"],
                                    [0.751111111, "red"],
                                    [1, "red"]],
                                opacity=0.5,
                                mapbox_style='open-street-map'
                                )
        fig.update_layout(margin=dict(b=0, t=0, l=0, r=0))
        if html_out is not None:
            fig.write_html(html_out)
        fig.show()

    def map_punctuality(self, time_stats: pd.DataFrame, html_out: str):
        m = folium.Map(location=(self.__warsaw_lat,
                       self.__warsaw_lon), zoom_start=14)

        data_dict = defaultdict(list)

        for busstop_name, line, lon, lat, arrived, expected, delay \
            in zip(time_stats['Busstop_name'], time_stats['Line'],
                   time_stats['Lon'], time_stats['Lat'],
                   time_stats['Arrived'], time_stats['Expected'],
                   time_stats['Delay [min]']):

            data_dict[busstop_name].append({'line': line, 'arrived': str(arrived),
                                            'lon': lon, 'lat': lat,
                                            'expected': str(expected),
                                            'delay': round(delay, 2)})

        marker_cluster = MarkerCluster().add_to(m)
        for busstop_name, val in data_dict.items():
            accumulator = str()
            for info in val:
                date, arrived_hour = info['arrived'].split(' ')
                excepted_hour = info['expected'].split(' ')[1]
                lon, lat = info['lon'], info['lat']

                accumulator += f"""<li>Bus line: {info['line']},
                           Arrival: {arrived_hour}, Expected: {excepted_hour},
                           Delay: {info['delay']} min</li>"""
            if accumulator == "":
                accumulator = "<li><\li>"
            popup_content = f"""
        <div style="width: 500px;">
        <h4> {busstop_name}, {date}</h4>
        <ul style="font-size: 14px;"> {accumulator}</ul>
        </div>
        """

            folium.Marker(location=[lat, lon], popup=popup_content,
                          icon=folium.Icon()).add_to(marker_cluster)
        
        
        m.save(html_out)
        webbrowser.open_new_tab(html_out)
    




from __future__ import annotations
from datetime import datetime as dt
from datetime import timedelta


def time_changer(time: str) -> dt:
    try:
        result = dt.strptime(time, '%Y-%m-%d %H:%M:%S')
        return result
    except ValueError:
        date_component, time_component = time.split(' ')
        date = dt.strptime(date_component, '%Y-%m-%d')
        time_sep = [int(x) for x in time_component.split(":")]
        delta = timedelta(
            hours=time_sep[0], minutes=time_sep[1], seconds=time_sep[2])
        return date + delta



from __future__ import annotations
import json
from typing import Set
import pandas as pd


from .modules.Bus import Bus
from .modules.SpeedAnalyzer import SpeedAnalyzer
from .modules.BusStopsManager import BusStopsManager
from .modules.RoutesManager import RoutesManager
from .modules.PunctualityAnalyzer import PunctualityAnalyzer
from .modules.TimetableManager import TimeTableMangager
from .modules.time_changer import time_changer


class BusStatistician:
  def __init__(self, bus_coordinates_file : str, timetable_file: str, routes_file : str, bus_stops_file : str):
    routes = RoutesManager(routes_file)
    bus_stops = BusStopsManager(bus_stops_file)
    timetable = TimeTableMangager(timetable_file)
    buses = self.__set_buses(bus_coordinates_file, timetable, routes)
    self.__speed_analyzer = SpeedAnalyzer(buses)
    self.__punctuality_analyzer = PunctualityAnalyzer(buses, bus_stops, routes)
    
    
  def __set_buses(self, buses_file : str, timetable : TimeTableMangager, routes : RoutesManager) -> Set[Bus]:
    with open(buses_file,'r', encoding='utf8') as file:
      temp_dict = json.load(file)
    
    buses : Set[Bus] = set()
    #register buses
    for bus_num in temp_dict["0"].keys():
      buses.add(Bus(bus_num)) 

    for all_buses in temp_dict.values():
      for bus in buses:
        params = all_buses.get(bus.get_vehicle_number())
        if params is not None: bus.update(params["Lines"], params["Brigade"],
                                          params["Lon"], params["Lat"],
                                          time_changer(params["Time"]))

    
    filtered_buses = []
    for bus in buses: 
      try:
        bus.set_timetable(timetable)
        filtered_buses.append(bus)
      except RuntimeError: pass

    filtered_buses = filter(lambda  x : x.number_of_states() > 1, filtered_buses)  
    buses = set(filtered_buses)
    
    for bus in buses: bus.determine_route(routes)
    return buses
  
 
  def get_speed_statistics(self) -> pd.DataFrame:
    df = self.__speed_analyzer.get_speed_statistics()
    return df
  
  
  
  def get_punctuality_statistics(self) -> pd.DataFrame:
    df = self.__punctuality_analyzer.compute_punctuality()
    return df
  


from __future__ import annotations
from typing import Set
import pandas as pd

from .Bus import Bus
from .BusStopsManager import BusStopsManager
from .RoutesManager import RoutesManager


class PunctualityAnalyzer:
    def __init__(self, buses: Set[Bus], bus_stops_manager: BusStopsManager,
                 routes_manager: RoutesManager) -> PunctualityAnalyzer:
        self.__buses = buses
        self.__bus_stops_manager = bus_stops_manager
        self.__routes_manager = routes_manager

    def compute_punctuality(self):
        counter = 0
        result = []
        bus_stop_name = self.__bus_stops_manager.get_bus_stop_name
        for bus in self.__buses:
            delay = bus.punctuality_compute(
                self.__bus_stops_manager, self.__routes_manager)
            counter += 1
            if counter % 100 == 0:
                print(f"{counter}/{len(self.__buses)} done")
            if delay:
                for busstop, list_of_info in delay.items():
                    coordinates = self.__bus_stops_manager.get_coordinates(
                        busstop)
                    for info in list_of_info:
                        result.append((bus.vehicle, info["Line"], busstop,
                                       bus_stop_name(busstop), coordinates.lon,
                                       coordinates.lat, info["Arrived"], info["Expected"]))

        df = pd.DataFrame(result, columns=['Bus_num', 'Line', 'Busstop_id',
                                           'Busstop_name', 'Lon', 'Lat',
                                           'Arrived', 'Expected'])
        df['Delay [min]'] = round((df['Arrived'] - df['Expected'])
                                  / pd.Timedelta(minutes=1), 2)

        return df



from __future__ import annotations
from datetime import datetime as dt
from datetime import timedelta
from timeit import default_timer as timer
from math import ceil
from collections import defaultdict
from typing import Dict, List, Tuple, DefaultDict

from .Coordinates import Coordinates
from .BusStopsManager import BusStopsManager
from .RoutesManager import RoutesManager
from .TimetableManager import TimeTableMangager


class Bus:
    def __init__(self, vehicle_num: str):
        self.__vehicle = vehicle_num
        self.__lines: List[str] = []
        self.__brigades: List[str] = []
        self.__timetable: Dict[str, Dict[str, List[Dict[str, str]]]] = {}
        self.__route_path: Dict[str, Tuple[str, List[dt]]] = {}
        self.__state: List[Tuple[dt, Coordinates]] = []
        self.__route = None
        self.__passed_station: str = None
        self.__current_state: int = 0

    @property
    def vehicle(self):
        return self.__vehicle
    
    # the following 3 methods are mainly for unit testing purposes
    def get_all_states(self):
        return tuple(self.__state)
    
    def get_all_lines(self):
        return tuple(self.__lines)
    
    def get_all_brigades(self):
        return tuple(self.__brigades)

    def __hash__(self) -> int:
        return hash(self.__vehicle)

    def __eq__(self, other):
        return self.__vehicle == other.__vehicle

    def update(self, line: str, brigade: str, longtitude: float, latitude: float, time: dt):
        # I do not add the same time
        if not self.__state or self.__state[-1][0] != time:
            coor = Coordinates(longtitude, latitude)
            
            ok_brigade = (brigade != " " and brigade != "")
            try:
                true_brigade = brigade if ok_brigade else self.__brigades[-1]
            except IndexError: return
            self.__brigades.append(true_brigade)
            self.__state.append((time, coor))
            self.__lines.append(line)
            

    def set_timetable(self, timetable_manager: TimeTableMangager) -> None:
        for line, brigade in zip(self.__lines, self.__brigades):
            key = line + "," + brigade
            # I connect line and brigade to get the unique key
            if line not in self.__timetable:
                try:
                    self.__timetable[key] = timetable_manager.get_schedule(
                        line, brigade)
                except KeyError:
                    raise RuntimeError("No suitable timetable found")

    def __get_current_timetable(self):
        return self.__timetable[self.line + "," + self.brigade]

    def get_vehicle_number(self) -> str:
        return self.__vehicle

    def number_of_states(self):
        return len(self.__state)

    def __str__(self) -> str:
        return str(self.__state)

    def __get_next_state(self):
        """sometimes in the data we have 2 different times but the same coordinates
        when a bus is stationery
         this function returns the first state that had change in its position."""
        current_coordinates = self.__state[self.__current_state][1]
        for i in range(self.__current_state + 1, len(self.__state)):
            next_state_coor = self.__state[i][1]
            if next_state_coor - current_coordinates != 0:
                return self.__state[i]
        return self.__state[self.__current_state]

    @property
    def line(self) -> str:
        return self.__lines[self.__current_state]

    @property
    def brigade(self) -> str:
        return self.__brigades[self.__current_state]

    def speed_compute(self) -> list[Tuple[int, Coordinates, str, str]]:
        result = []
        for past, future in zip(self.__state, self.__state[1:]):
            time: float = (future[0] - past[0]).total_seconds()

            if time != 0:
                # 3.6 is conversion m/s to km/h
                speed = ceil(3.6 * (future[1]-past[1])/time)
                result.append((speed, past[1], self.__vehicle, self.line))
        return result

    def determine_route(self, routes: RoutesManager) -> None:
        some_list = []
        def time_difference(entry): return abs(
            (entry['Time'] - self.__state[self.__current_state][0]).total_seconds())
        for time_entry in self.__get_current_timetable().values():
            some_list.append(min(time_entry, key=time_difference))

        if not some_list:
            return
        min_time = min(some_list, key=time_difference)
        self.__route = min_time["Route"]
        path = routes.get_route(self.line, self.__route)

        for num, busstop in path.items():
            times = self.__get_current_timetable().get(busstop)
            if times is None:
                times = []
            self.__route_path[num] = (busstop, [x['Time'] for x in times])

    def __determine_passed_stop(self, busstops: BusStopsManager) -> None:
        if not self.__route_path or self.__current_state >= len(self.__state) - 1:
            return
        temp_list = []
        for index, busstop_with_time in self.__route_path.items():
            busstop = busstop_with_time[0]
            try:
                distance = self.__state[self.__current_state][1] - \
                    busstops.get_coordinates(busstop)
            except KeyError:
                # Sometimes there's a problem with busstops that are suspended
                # but it is still present in the routes.
                continue
            temp_list.append((index, distance))

        # Sorting is O(N log N) but here the data is restricted to max 20-30 values so doing
        # O(N) will take more
        two_min_values = sorted(temp_list, key=lambda x: x[1])[0:2]

        # Make the first busstop the former one
        if int(two_min_values[0][0]) > int(two_min_values[1][0]):
            two_min_values.reverse()

        future_distances = [(index, busstops.get_coordinates(
                            self.__route_path[index][0]) - self.__get_next_state()[1])
                            for index, _ in two_min_values]

        """
    Four cases:
    1. Distance to the former one increased, the latter one decreased -> bus heads to the latter one
    2. Distance to the former one decreased, the latter one increased -> bus heads to the latter one
    3. Two distances decreased -> bus goes to the former bus stop
    4. Two distance increased bus passed the later bus stop
        """
        def helper(index):
            return future_distances[index][1] - two_min_values[index][1]

        if helper(0) > 0:
            if helper(1) < 0: self.__passed_station = two_min_values[0][0]
            else: self.__passed_station = two_min_values[1][0]
        else: self.__passed_station = str(int(two_min_values[0][0]) - 1)

    def punctuality_compute(self, busstops: BusStopsManager,
                            routes: RoutesManager) -> DefaultDict[str, List[Dict[str, str]]]:
        self.__determine_passed_stop(busstops)
        meters_to_stop = 10
        offset_in_meters = 5
        delay: Dict[str, List[Dict[str, str]]] = defaultdict(list)
        if self.__passed_station is None:
            return delay
        passed_stop = int(self.__passed_station)+1

        prev_line = self.line
        prev_brigade = self.brigade
        check_for_route = False

        while self.__current_state < len(self.__state) - 1:
            if passed_stop >= len(self.__route_path) or check_for_route \
                    or prev_line != self.line \
                    or prev_brigade != self.brigade:
                check_for_route = False
                self.__current_state += 1
                self.determine_route(routes)
                prev_line = self.line
                prev_brigade = self.brigade
                passed_stop = int(self.__passed_station)+1
                continue

            present_time, present_coord = self.__state[self.__current_state]
            busstop, arrival_times = self.__route_path[f"{passed_stop}"]
            if not arrival_times:
                self.__current_state += 1
                self.__determine_passed_stop(busstops)
                passed_stop = int(self.__passed_station)+1
                continue

            try:
                # sometimes there is a mismatch between the busstops for the day
                # When a busstop is suspended but a route still has it. It is pretty rare though
                busstop_coord = busstops.get_coordinates(busstop)
            except KeyError:
                continue
            present_distance = present_coord - busstop_coord
            if present_distance < meters_to_stop:  # We consider that the bus arrived at the bus stop
                passed_stop += 1
                self.__current_state += 1
                time_diff = [(present_time - arrival_time).total_seconds()
                             for arrival_time in arrival_times]
                min_time_diff = min(time_diff, key=lambda x: abs(x))
                delay[busstop].append({"Line": self.line, "Arrived": present_time,
                                      "Expected": arrival_times[time_diff.index(min_time_diff)]})
                continue

            future_time, future_coord = self.__state[self.__current_state+1]
            if future_coord == present_coord:
                self.__current_state += 1
                continue
            future_distance = future_coord - busstop_coord

            if future_distance < meters_to_stop or future_distance - offset_in_meters < present_distance:
                self.__current_state += 1
                continue

            check_for_further_distance = False
            for j in range(self.__current_state + 2, min(self.__current_state + 10, len(self.__state))):
                far_future_coor = self.__state[j][1]
                if far_future_coor - busstop_coord < present_distance - offset_in_meters:
                    self.__current_state = j
                    check_for_further_distance = True
                    break
            if check_for_further_distance:
                continue

            if (future_time - present_time).total_seconds() > 60:  # Too big gap in time
                self.__current_state += 1
                self.__determine_passed_stop(busstops)
                passed_stop = int(self.__passed_station)+1
                continue

            # Case: present_distance > future_distance means that we gone past the stop
            # and we are more than 10 m from it
            # We do some rough approximations that includes speed
            # Calculate the speed between these two points, find the time required to get to the stop
            # add to the present_time and compare with the expected time
            if future_distance < 500:
                self.__current_state += 1
                passed_stop += 1
                time_diff = (future_time - present_time).total_seconds()
                # in m/s:
                speed1 = ceil((future_coord - present_coord) / time_diff) + 1
                speed2 = ceil((future_distance - present_distance) / time_diff) + 1  
                
                # speed2 better represent a situation where a bus passed the
                # latter bus stop but it may give unrealistic result,
                # hence I impose a treshhold
                speed = speed1 if speed2 > 50 else speed2

                time_to_stop = timedelta(seconds=int((present_distance / speed)))
                time_of_arrival = present_time + time_to_stop

                time_diff = [(time_of_arrival - arrival_time).total_seconds()
                             for arrival_time in arrival_times]
                min_time_diff = min(time_diff, key=lambda x: abs(x))
                delay[busstop].append({"Line": self.line, "Arrived": time_of_arrival,
                                      "Expected": arrival_times[time_diff.index(min_time_diff)]})
            else:
                check_for_route = True

        return delay



import json


class RoutesManager:
    def __init__(self, filename: str):
        with open(filename, 'r', encoding='utf8') as file:
            self.__routes = json.load(file)

    def get_route(self, line: str, route_name: str):
        return self.__routes[line][route_name]



import setuptools
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ztm_buses",  
    version="0.0.1",
    author="Maksym Matviienko", 
    author_email="xxxmax337@example.com",
    description="Warsaw ZTM API Fetcher and Data Analyzer",
    long_description=long_description,
    install_requires=["requests","urllib3", "folium","plotly_express", "pandas"],
    setup_requires=["requests","urllib3","folium", "plotly_express", "pandas"],
    long_description_content_type="text/markdown",
    url="https://github.com/MMax337/ztm_buses/tree/main",
    packages=setuptools.find_packages(exclude=["contrib", "docs", "tests*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from __future__ import annotations
from typing import KeysView
import json

from .Coordinates import Coordinates


class BusStopsManager:
    def __init__(self, filename: str) -> BusStopsManager:
        self.__bus_stops = {}
        self.__bus_stop_names = {}
        with open(filename, 'r', encoding='utf8') as file:
            busstop_coordinates = json.load(file)

        for busstop, val in busstop_coordinates.items():
            # busstop is string of the form "XXXX,XX"
            coord = val["Lon,Lat"]
            name = val["stop_name"]
            bus_complex_id = busstop.split(',')[0]
            self.__bus_stops[busstop] = Coordinates(
                float(coord[0]), float(coord[1]))
            if bus_complex_id not in self.__bus_stop_names:
                self.__bus_stop_names[bus_complex_id] = name

    def get_coordinates(self, busstop: str):
        return self.__bus_stops[busstop]

    def get_bus_nums(self) -> KeysView[str]:
        return self.__bus_stops["0"].keys()

    def get_bus_stop_name(self, busstop: str) -> str:
        # busstop is string of the form "XXXX,XX"
        bus_complex_id, bus_stop_no = busstop.split(',')
        return self.__bus_stop_names[bus_complex_id] + " " + bus_stop_no



from __future__ import annotations
from typing import Set, List
import pandas as pd

from .Bus import Bus
from .RoutesManager import RoutesManager
from .BusStopsManager import BusStopsManager


class SpeedAnalyzer:
    def __init__(self, buses: Set[Bus]) -> SpeedAnalyzer:
        self.__buses = buses

    def __speed_compute(self) -> List[float]:
        speeds: List = []
        for bus in self.__buses:
            val = bus.speed_compute()
            if val:
                speeds += val
        return speeds

    def get_speed_statistics(self):
        speeds: List[float] = self.__speed_compute()
        speeds_df_data = [(speed, coord.lon, coord.lat, vehicle_num, line_num)
                          for speed, coord, vehicle_num, line_num in speeds]
        speeds_df = pd.DataFrame(speeds_df_data, columns=[
                                 'speed', 'lon', 'lat', 'vehicle_number', 'line_number'])

        return speeds_df
